import click
import yaml
from pathlib import Path
from base64 import b64encode

from . import UOTP
from .packet import TimeRequest
from .util import OTPUtil


config = {}
fp = None


def save_config():
    fp.seek(0)
    yaml.dump(config, fp, default_flow_style=False)


@click.group(invoke_without_command=True)
@click.option('--conf', default='~/.config/uotp/config.yml', envvar='UOTP_CONF', help='Path to the configuration file.')
@click.pass_context
def cli(ctx, conf):
    global config, fp

    path = Path(conf).expanduser()
    path.parent.mkdir(parents=True, exist_ok=True)
    if path.exists():
        fp = click.open_file(str(path), 'r+', encoding='utf-8')
        config = yaml.load(fp)
    else:
        fp = click.open_file(str(path), 'w', encoding='utf-8')
        config = {
            'account': None,
            'timediff': 0,
        }
        save_config()
        click.echo('A new configuration file has been created on `{}`.'.format(path))
        click.echo()

    if ctx.invoked_subcommand is None:
        ctx.invoke(get)


@cli.command()
@click.pass_context
def new(ctx):
    if config['account']:
        click.confirm('Account already exists. Do you want to replace it?', abort=True)

    ctx.invoke(sync)

    uotp = UOTP()
    uotp.issue_account()

    config['account'] = {
        'oid': uotp.account_oid,
        'seed': uotp.account_seed,
        'serial_number': uotp.account_serial_number,
        'user_hash': uotp.account_id.encode(),
    }
    save_config()

    click.echo('A new account has been issued.')
    click.echo('Please keep your configuration file safe as it is not possible to recover the account if it gets lost.')
    click.echo()
    click.echo('Serial Number: {}'.format(uotp.account_serial_number))


@cli.command()
def sync():
    time = TimeRequest()()['time']
    timediff = time - OTPUtil.now()

    config['timediff'] = timediff
    save_config()

    click.echo('Time synchronized with the remote server (offset: {}sec).'.format(timediff))


@cli.command()
@click.option('-s', '--autosync', is_flag=True, help='Automatically synchronize time before generating OTP token')
@click.pass_context
def get(ctx, autosync):
    if not config['account']:
        if click.confirm('Account not exists. Do you want to issue one now?', default=True):
            ctx.invoke(new)
        else:
            click.echo('Please issue a new account first. You can do this with `uotp new`.')
        return

    if autosync:
        ctx.invoke(sync)

    uotp = UOTP(
        config['account']['user_hash'].decode(),
        config['account']['oid'],
        config['account']['seed'],
        config['account']['serial_number']
    )
    uotp.sync_time(config['timediff'])
    token = uotp.generate_token()

    token = OTPUtil.humanize(token, char=' ', each=3, maxgroup=2)
    click.echo('OTP Token: {}'.format(token))


@cli.command()
def info():
    if not config['account']:
        click.echo('Please issue a new account first. You can do this with `uotp new`.')
        return

    click.echo('S/N: {}'.format(config['account']['serial_number']))
    click.echo('Time offset: {}sec'.format(config['timediff']))
    click.echo('Oid: {}'.format(config['account']['oid']))
    click.echo('Seed: {}'.format(b64encode(config['account']['seed']).decode()))


@cli.command()
def gui():
    import wx
    from .gui import MainWindow
    app = wx.App()
    MainWindow()
    app.MainLoop()


if __name__ == '__main__':
    cli()



from .help import HelpRequest
from .information import InformationRequest
from .issue import IssueRequest
from .reset_error_count import ResetErrorCountRequest
from .time import TimeRequest
from .use_history import UseHistoryRequest



from struct import unpack_from

from .base import Packet, Opcode


class InformationRequest(Packet):
    OPCODE = Opcode.Information

    @classmethod
    def _encode_payload(cls, data: dict) -> bytes:
        return b''

    @classmethod
    def _decode_payload(cls, payload: bytes) -> dict:
        oid, seed, partner = map(lambda x: x.strip(), unpack_from('!11s40s80s', payload, 0))

        return {
            'oid': int(oid),
            'seeed': seed,
            'partner': partner,
        }



from functools import reduce
from hashlib import sha1

from .util import OTPUtil


class OTPTokenGenerator:
    def __init__(self, oid: int, seed: bytes):
        self.__oid = oid
        self.__seed = seed
        self.__timediff = 0

    def compensate_time_deviation(self, timediff: int):
        self.__timediff = timediff

    def generate_token(self) -> str:
        now = OTPUtil.now()
        now += self.__timediff

        time = now // 10
        oid = self.__oid

        acc_seed = bytearray(11)
        time_seed = bytearray(11)

        for i in reversed(range(11)):
            acc_seed[i] = int(oid) & 255
            time_seed[i] = int(time) & 255

            oid >>= 8
            time >>= 8

        acc_seed += self.__seed

        acc_seed_0 = bytearray(64)
        acc_seed_0[:20] = sha1(acc_seed).digest()

        acc_seed_1 = acc_seed_0[:]
        acc_seed_2 = acc_seed_0[:]

        for i in range(64):
            acc_seed_1[i] ^= 54
            acc_seed_2[i] ^= 92

        digest = sha1(acc_seed_1 + time_seed).digest()
        digest = sha1(acc_seed_2 + digest).digest()

        digit = digest[-1] & 0xf

        # noinspection PyShadowingNames
        token = reduce(lambda r, i: (r << 8) | digest[digit + i], range(4))
        token &= 0xffffdb

        rem = now % 30
        if 10 <= rem < 20:
            token |= 4
        elif 20 <= rem < 30:
            token |= 32

        token = token % 10000000
        token = str(token).zfill(7)

        return token



from datetime import datetime
from struct import unpack_from

from .base import Packet, Opcode


class UseHistoryRequest(Packet):
    OPCODE = Opcode.UseHistory

    @classmethod
    def _encode_payload(cls, data: dict) -> bytes:
        # [period] 1: 3 days / 2: 1 week / 3: 1 month
        return '{:0>4}{: <1}'.format(
            data['page'], data['period']
        ).encode()

    @classmethod
    def _decode_payload(cls, payload: bytes) -> dict:
        date_start, date_end, = map(lambda x: datetime.strptime(x.decode(), '%Y-%m-%d'), unpack_from('!10s10s', payload, 0))
        page_total, page_current, data_count = map(lambda x: int(x), unpack_from('!4s4s2s', payload, 20))

        entries = []
        offset = 30
        for i in range(data_count):
            date, time, typ, name = unpack_from('!10s8s40s40s', payload, offset)
            entries.append({
                'datetime': datetime.strptime((date + time).decode(), '%Y-%m-%d%H:%M:%S'),
                'type': typ.decode('cp949').strip(),
                'name': name.decode('cp949').strip(),
            })

            offset += 98

        return {
            'period': {
                'start': date_start,
                'end': date_end,
            },
            'page': {
                'total': page_total,
                'current': page_current,
            },
            'entries': entries,
        }



from datetime import datetime
from math import ceil


class OTPUtil:
    @staticmethod
    def now() -> int:
        now = datetime.now()
        return (now.year - 2000) * 31536000 \
            + (now.month - 1) * 2592000 \
            + (now.day - 1) * 86400 \
            + now.hour * 3600 \
            + now.minute * 60 \
            + now.second

    @staticmethod
    def humanize(text: str, char: str, each: int, maxgroup: int = -1) -> str:
        if maxgroup == -1:
            maxgroup = ceil(len(text) / each)
        r = char.join(text[i:i + each] for i in range(0, len(text), each) if i / each < maxgroup)
        r += text[each * maxgroup:]
        return r


class Int32:
    MAX = 0x7fffffff
    MIN = -0x7fffffff

    def __init__(self, value):
        value = int(value)
        if not (self.MIN <= value <= self.MAX):
            value = (value + self.MAX + 1) % (2 * (self.MAX + 1)) - self.MAX - 1
        self.value = value

    def __str__(self):
        return str(self.value)

    def __repr__(self):
        return '<Int32: {}>'.format(self.value)

    def __format__(self, format_spec):
        return format(self.value, format_spec)

    def __int__(self):
        return self.value

    def __add__(self, other):
        return Int32(self.value + other)

    def __sub__(self, other):
        return Int32(self.value - other)

    def __and__(self, other):
        return Int32(self.value & other)

    def __lshift__(self, other):
        return Int32(self.value << other)

    def __rshift__(self, other):
        return Int32(self.value >> other)

    def __xor__(self, other):
        return Int32(self.value ^ other)

    def __radd__(self, other):
        return Int32(other + self.value)

    def __rsub__(self, other):
        return Int32(other - self.value)

    def __rxor__(self, other):
        return Int32(other ^ self.value)

    def to_bytes(self, length=4, byteorder='little', signed=True):
        return self.value.to_bytes(length, byteorder, signed=signed)




from .packet import IssueRequest, ResetErrorCountRequest, TimeRequest, UseHistoryRequest
from .token import OTPTokenGenerator, OTPUtil


class UOTP:
    def __init__(self, _id: str = '', oid: int = None, seed: bytes = None, serial_number: str = None):
        self.__account = {
            'id': _id.encode(),
            'oid': oid,
            'seed': seed,
            'serial_number': serial_number,
        }
        self.__generator = OTPTokenGenerator(self.account_oid, self.account_seed)
        self.__timediff = 0

    def generate_token(self) -> str:
        self.__generator.compensate_time_deviation(self.__timediff)
        return self.__generator.generate_token()

    def sync_time(self, timediff=None) -> bool:
        try:
            if timediff is None:
                time = TimeRequest()()['time']
                timediff = time - OTPUtil.now()
        except RuntimeError:
            return False
        else:
            self.__timediff = timediff
            return True

    def issue_account(self) -> bool:
        try:
            resp = IssueRequest()()
            self.__account = {
                'id': resp['user_hash'],
                'oid': resp['oid'],
                'seed': resp['seed'],
                'serial_number': OTPUtil.humanize(resp['serial_number'], char='-', each=4)
            }
        except RuntimeError:
            return False
        else:
            self.__generator = OTPTokenGenerator(self.account_oid, self.account_seed)
            return True

    def reset_error(self) -> bool:
        try:
            req = ResetErrorCountRequest()
            req.set_oid(self.account_oid)
            req.set_encryption_info(self.__account['id'], self.generate_token())
            req()
        except RuntimeError:
            return False
        else:
            return True

    def get_history(self, page=1) -> dict:
        req = UseHistoryRequest()
        req.set_oid(self.account_oid)
        req.set_encryption_info(self.__account['id'], self.generate_token())
        req['page'] = page
        req['period'] = 3
        return req().params

    @property
    def account_id(self):
        return self.__account['id'].decode()

    @property
    def account_oid(self):
        return self.__account['oid']

    @property
    def account_seed(self):
        return self.__account['seed']

    @property
    def account_serial_number(self):
        return self.__account['serial_number']



from .base import Packet, Opcode


class ResetErrorCountRequest(Packet):
    OPCODE = Opcode.ResetErrorCount

    @classmethod
    def _encode_payload(cls, data: dict) -> bytes:
        return b''

    @classmethod
    def _decode_payload(cls, payload: bytes) -> dict:
        return {}



from binascii import unhexlify
from struct import unpack

from .base import Packet, Opcode


class IssueRequest(Packet):
    OPCODE = Opcode.Issue

    def __init__(self):
        super().__init__()
        self.set_encryption_info(self.generate_shared_key())

    @classmethod
    def _encode_payload(cls, data: dict) -> bytes:
        return b''

    @classmethod
    def _decode_payload(cls, payload: bytes) -> dict:
        serial_number, oid, seed, user_hash, issue_info = map(lambda x: x.strip(), unpack('!20s11s40s64s80s', payload))

        return {
            'oid': int(oid),
            'seed': unhexlify(seed),
            'serial_number': serial_number.decode(),
            'user_hash': user_hash,
            'issue_info': issue_info,
        }



import wx
import yaml
from pathlib import Path

from .packet import IssueRequest, TimeRequest
from . import UOTP


class MainWindow(wx.Frame):
    CONFIG_PATH = '~/.config/uotp/config.yml'
    REFRESH_INTERVAL = 30

    def __init__(self):
        super().__init__(
            None,
            title='μOTP+',
            size=(320, 240),
            style=wx.CAPTION | wx.MINIMIZE_BOX | wx.CLOSE_BOX,
        )

        path = Path(self.CONFIG_PATH).expanduser()
        path.parent.mkdir(parents=True, exist_ok=True)
        if path.exists():
            self.config_fp = open(str(path), 'r+', encoding='utf-8')
            self.config = yaml.load(self.config_fp)
        else:
            self.config_fp = open(str(path), 'w', encoding='utf-8')
            self.config = {
                'account': None,
                'timediff': 0,
            }
            self.save_config()

        if not self.config['account']:
            self.issue()

        wrapper = wx.Panel(self)
        sizer_wrapper = wx.BoxSizer(wx.VERTICAL)

        self.wg_gauge_time = wx.Gauge(wrapper, range=self.REFRESH_INTERVAL, style=wx.GA_HORIZONTAL | wx.GA_SMOOTH)
        sizer_wrapper.Add(self.wg_gauge_time, flag=wx.EXPAND)

        panel = wx.Panel(wrapper)
        box = wx.BoxSizer(wx.VERTICAL)

        self.wg_txt_token = wx.TextCtrl(panel, style=wx.TE_READONLY | wx.TE_CENTRE | wx.BORDER_NONE)
        self.wg_txt_token.SetFont(wx.Font(32, wx.DEFAULT, wx.NORMAL, wx.BOLD))
        self.wg_txt_token.SetLabelText('INITIALIZING')
        box.Add(self.wg_txt_token, flag=wx.EXPAND)

        box.AddSpacer(60)

        box.Add(wx.StaticText(panel, label="S/N"), flag=wx.ALIGN_CENTRE)

        box.AddSpacer(5)

        self.wg_txt_serial = wx.TextCtrl(panel, style=wx.TE_READONLY | wx.TE_CENTRE | wx.BORDER_NONE)
        self.wg_txt_serial.SetLabelText(self.config['account']['serial_number'])
        box.Add(self.wg_txt_serial, flag=wx.ALIGN_CENTRE)

        panel.SetSizer(box)
        sizer_wrapper.Add(panel, proportion=1, flag=wx.ALL | wx.EXPAND, border=20)
        wrapper.SetSizer(sizer_wrapper)

        self.uotp = UOTP(
            self.config['account']['user_hash'].decode(),
            self.config['account']['oid'],
            self.config['account']['seed'],
            self.config['account']['serial_number']
        )
        self.uotp.sync_time()

        self.timeout = -1
        self.timer = wx.Timer(self)
        self.Bind(wx.EVT_TIMER, self.on_update, self.timer)
        self.timer.Start(1000)

        self.on_update()

        self.Centre()
        self.Show()

    def issue(self):
        req = IssueRequest()
        req['mno'] = 'KTF'
        req['hw_id'] = 'GA15'
        req['hw_model'] = 'SM-N900P'
        req['version'] = (2, 0)

        resp = req()
        self.config['account'] = resp.params
        self.save_config()

    def save_config(self):
        self.config_fp.seek(0)
        yaml.dump(self.config, self.config_fp, default_flow_style=False)

    def on_update(self, event=None):
        self.timeout -= 1
        if self.timeout < 0:
            self.timeout = self.REFRESH_INTERVAL
            self.wg_txt_token.SetValue(self.uotp.generate_token())
        self.wg_gauge_time.SetValue(self.timeout)


if __name__ == '__main__':
    app = wx.App()
    MainWindow()
    app.MainLoop()



from .base import Packet, Opcode


class HelpRequest(Packet):
    OPCODE = Opcode.Help

    @classmethod
    def _encode_payload(cls, data: dict) -> bytes:
        return b''

    @classmethod
    def _decode_payload(cls, payload: bytes) -> dict:
        payload = payload[:-8].decode('cp949')
        return {
            'messages': payload.split('|'),
        }



from struct import unpack

from .base import Packet, Opcode


class TimeRequest(Packet):
    OPCODE = Opcode.Time
    SIMPLE = True

    @classmethod
    def _encode_payload(cls, data: dict) -> bytes:
        return b''

    @classmethod
    def _decode_payload(cls, payload: bytes) -> dict:
        time, = unpack("!I", payload)

        return {
            'time': time
        }



from setuptools import setup, find_packages


with open('README.rst', 'r', encoding='utf-8') as f:
    long_description = f.read()


setup(
    name='uotp',
    version='0.4.0',
    description='μOTP+: The next generation OTP toolkit',
    long_description=long_description,

    packages=find_packages(),

    author='Bae Junehyeon',
    author_email='devunt' '@' 'gmail.com',
    url='https://github.com/devunt/uotp',
    download_url='',
    license='Public Domain',

    python_requires='~=3.4',
    install_requires=[
        'click',
        'PyYAML',
        'wxPython',
    ],
    entry_points={
        'console_scripts': [
            'uotp = uotp.cli:cli',
        ]
    },

    keywords=['otp'],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'License :: Public Domain',
        'Natural Language :: Korean',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3 :: Only',
        'Topic :: Security',
        'Topic :: Security :: Cryptography',
        'Topic :: Utilities',
    ],
)



from binascii import unhexlify
from enum import Enum, unique
from hashlib import sha1, sha256
from random import choice
from socket import socket
from struct import pack, unpack

from ..seed import SEED
from ..util import OTPUtil


@unique
class Status(Enum):
    OK = '0000'
    C002 = 'C002'
    S003 = 'S003'
    Error = -1


@unique
class Opcode(Enum):
    Error = -1
    Information = 402
    Time = 407
    Issue = 451
    ResetErrorCount = 452
    UseHistory = 453
    Help = 454


class Packet:
    OPCODE = None
    SIMPLE = False

    def __init__(self):
        self.status = Status.OK
        self.params = {}

        self.oid = ''

        self.shared_crypto_key = b''
        self.extra_token = b''

    def __call__(self):
        crypto_key = self.__get_crypto_key()

        s = socket()
        s.connect(('211.49.97.230', 20004))
        s.send(self.__to_bytes(crypto_key))

        s.recv(1)
        size = int(s.recv(5))
        data = s.recv(size)

        s.close()

        return self.__from_bytes(data, crypto_key)

    def __getitem__(self, item):
        return self.params[item]

    def __setitem__(self, key, value):
        self.params[key] = value

    def __repr__(self):
        return '<Packet: {}>'.format(self.OPCODE)

    def set_oid(self, oid):
        self.oid = oid

    def set_encryption_info(self, shared_crypto_key: bytes = b'', extra_token: str = b''):
        if shared_crypto_key:
            self.shared_crypto_key = shared_crypto_key
        if extra_token:
            self.extra_token = '{:0>7} '.format(extra_token).encode()

    def __get_crypto_key(self) -> bytes:
        if self.shared_crypto_key:
            if self.extra_token:
                hasher = sha1()
                hasher.update(unhexlify(self.shared_crypto_key))
                hasher.update(self.extra_token)
                return hasher.digest()
            else:
                return self.shared_crypto_key
        else:
            return b''

    def __generate_payload_common_header(self) -> bytes:
        return '{: <3}{: <11}{: <16}{: <4}{:0>4}{:0>4}'.format(
            choice(('KTF', 'SKT', 'LGT')), self.oid,
            choice(('SM-G950S', 'SM-G955L', 'SM-G920K')), 'GA15', 2, 0
        ).encode()

    def __to_bytes(self, crypto_key: bytes = b'') -> bytes:
        payload = self._encode_payload(self.params)

        if not self.SIMPLE:
            payload = self.__generate_payload_common_header() + payload

        payload += self.extra_token

        if crypto_key and payload:
            payload = SEED.encrypt(crypto_key, payload)

        shared_key = self.shared_crypto_key.rjust(64, b' ')
        codes = '{:0>4}{:0>3}'.format(self.status.value, self.OPCODE.value).encode()

        body = shared_key + codes + payload
        header = 'S{:0>5}'.format(len(body)).encode()

        return header + body

    @classmethod
    def __from_bytes(cls, data: bytes, crypto_key: bytes = b'') -> 'Packet':
        self = Packet()

        shared_key, status, opcode = unpack('64s4s3s', data[:71])

        self.shared_crypto_key = unhexlify(shared_key.decode().rstrip(' '))

        try:
            self.status = Status(status.decode())
        except ValueError:
            self.status = Status.Error

        try:
            self.OPCODE = Opcode(int(opcode.decode()))
        except ValueError:
            self.OPCODE = Opcode.Error

        payload = data[71:]

        if self.shared_crypto_key or crypto_key:
            payload = SEED.decrypt(self.shared_crypto_key or crypto_key, payload)

        if self.status is not Status.OK:
            raise RuntimeError(self.status, payload.decode('euc-kr'))

        self.params = cls._decode_payload(payload)

        return self

    @classmethod
    def _encode_payload(cls, data: dict) -> bytes:
        raise NotImplementedError

    @classmethod
    def _decode_payload(cls, payload: bytes) -> dict:
        raise NotImplementedError

    @staticmethod
    def generate_shared_key() -> bytes:
        return sha256(pack('!I', OTPUtil.now())).hexdigest().encode()
